from typing import Any, Iterator, TypeAlias

from ..base import BaseSolver

Node: TypeAlias = tuple[int, int]
Edge: TypeAlias = tuple[Node, Node]


class Solver(BaseSolver):
    def solve(self, input: str) -> Iterator[Any]:
        grid = input.splitlines()
        n_rows, n_cols = len(grid), len(grid[0])

        marked = {(row, col): False for row in range(n_rows) for col in range(n_cols)}

        total_1: int = 0
        total_2: int = 0

        while not all(marked.values()):
            row, col = next(k for k, v in marked.items() if not v)

            cells: set[tuple[int, int]] = set()
            perimeter: int = 0
            value = grid[row][col]
            queue: list[tuple[int, int]] = [(row, col)]

            edges: set[Edge] = set()
            sides: list[tuple[Edge, ...]] = []

            while queue:
                row, col = queue.pop(0)

                if marked[row, col]:
                    continue

                cells.add((row, col))
                marked[row, col] = True

                if row == 0 or grid[row - 1][col] != value:
                    perimeter += 1

                    if ((row, col), (row - 1, col)) not in edges:
                        side: list[Edge] = [(((row, col), (row - 1, col)))]
                        for col2 in list(range(col - 1, -1, -1)):
                            if grid[row][col2] != value or (
                                row > 0 and grid[row][col2] == grid[row - 1][col2]
                            ):
                                break
                            side.append((((row, col2), (row - 1, col2))))
                        for col2 in range(col + 1, n_cols):
                            if grid[row][col2] != value or (
                                row > 0 and grid[row][col2] == grid[row - 1][col2]
                            ):
                                break
                            side.append((((row, col2), (row - 1, col2))))
                        sides.append(tuple(side))
                        edges = edges.union(side)
                else:
                    queue.append((row - 1, col))

                if col == 0 or grid[row][col - 1] != value:
                    perimeter += 1

                    if ((row, col), (row, col - 1)) not in edges:
                        side: list[Edge] = [(((row, col), (row, col - 1)))]
                        for row2 in range(row - 1, -1, -1):
                            if grid[row2][col] != value or (
                                col > 0 and grid[row2][col] == grid[row2][col - 1]
                            ):
                                break
                            side.append((((row2, col), (row2, col - 1))))
                        for row2 in range(row + 1, n_rows):
                            if grid[row2][col] != value or (
                                col > 0 and grid[row2][col] == grid[row2][col - 1]
                            ):
                                break
                            side.append((((row2, col), (row2, col - 1))))
                        sides.append(tuple(side))
                        edges = edges.union(side)
                else:
                    queue.append((row, col - 1))

                if row + 1 == n_rows or grid[row + 1][col] != value:
                    perimeter += 1

                    if ((row, col), (row + 1, col)) not in edges:
                        side: list[Edge] = [(((row, col), (row + 1, col)))]
                        for col2 in list(range(col - 1, -1, -1)):
                            if grid[row][col2] != value or (
                                row + 1 < n_rows
                                and grid[row][col2] == grid[row + 1][col2]
                            ):
                                break
                            side.append((((row, col2), (row + 1, col2))))
                        for col2 in range(col + 1, n_cols):
                            if grid[row][col2] != value or (
                                row + 1 < n_rows
                                and grid[row][col2] == grid[row + 1][col2]
                            ):
                                break
                            side.append((((row, col2), (row + 1, col2))))
                        sides.append(tuple(side))
                        edges = edges.union(side)
                else:
                    queue.append((row + 1, col))

                if col + 1 == n_cols or grid[row][col + 1] != value:
                    perimeter += 1

                    if ((row, col), (row, col + 1)) not in edges:
                        side: list[Edge] = [(((row, col), (row, col + 1)))]
                        for row2 in range(row - 1, -1, -1):
                            if grid[row2][col] != value or (
                                col + 1 < n_cols
                                and grid[row2][col] == grid[row2][col + 1]
                            ):
                                break
                            side.append((((row2, col), (row2, col + 1))))
                        for row2 in range(row + 1, n_rows):
                            if grid[row2][col] != value or (
                                col + 1 < n_cols
                                and grid[row2][col] == grid[row2][col + 1]
                            ):
                                break
                            side.append((((row2, col), (row2, col + 1))))
                        sides.append(tuple(side))
                        edges = edges.union(side)
                else:
                    queue.append((row, col + 1))

            self.logger.info(
                f"region with {value}: {len(cells)} * {perimeter} = {len(cells) * perimeter}, {len(cells)} * {len(sides)} = {len(cells) * len(sides)}"
            )
            total_1 += len(cells) * perimeter
            total_2 += len(cells) * len(sides)

        yield total_1
        yield total_2
